package cn.sliyun.dotwallet;

import cn.sliyun.dotwallet.entity.response.data.SignParam;
import cn.sliyun.dotwallet.util.Util;
import com.subgraph.orchid.encoders.Hex;
import org.bitcoinj.core.*;
import org.bitcoinj.crypto.ChildNumber;
import org.bitcoinj.crypto.DeterministicKey;
import org.bitcoinj.crypto.HDKeyDerivation;
import org.bitcoinj.script.Script;
import org.bitcoinj.script.ScriptBuilder;

import java.util.ArrayList;
import java.util.List;

public class BSVTransaction extends Transaction {
    public static final byte ForkID = 0x40;
    public BSVTransaction(NetworkParameters params, byte[] payloadBytes) throws ProtocolException {
        super(params, payloadBytes);
    }

    public BSVTransaction(NetworkParameters params, String rawTx) throws ProtocolException {
        super(params, Hex.decode(rawTx));
    }

    private int appendVarInt(ArrayList<byte[]> byteArray, byte[] connectedScript) {
        long len = connectedScript.length;
        int scriptLen = 0;
        if (len < 0xfd) {
            scriptLen += appendBytes(byteArray, Util.byteToBytesLittle((byte) len));
            return scriptLen;
        }

        if (len <= Short.MAX_VALUE) {
            scriptLen += appendBytes(byteArray, Util.byteToBytesLittle((byte) 0xfd));
            scriptLen += appendBytes(byteArray, Util.shortToBytesLittle((short) len));
            return scriptLen;
        }

        if (len <= Integer.MAX_VALUE) {
            scriptLen += appendBytes(byteArray, Util.byteToBytesLittle((byte) 0xfe));
            scriptLen += appendBytes(byteArray, Util.intToBytesLittle((int) len));
            return scriptLen;
        }

        scriptLen += appendBytes(byteArray, Util.byteToBytesLittle((byte) 0xff));
        scriptLen += appendBytes(byteArray, Util.longToBytesLittle(len));
        return scriptLen;
    }

    private int appendBytes(ArrayList<byte[]> byteArray, byte[] newByte) {
        byteArray.add(newByte);
        return newByte.length;
    }

    private byte[] getFinalBytes(ArrayList<byte[]> byteArray, int totalLen) {
        byte[] finalBytes = new byte[totalLen];
        int off = 0;
        for (byte[] tmpBytes : byteArray) {
            System.arraycopy(tmpBytes, 0, finalBytes, off, tmpBytes.length);
            off += tmpBytes.length;
        }
        return finalBytes;
    }

    private Sha256Hash hashForSignature(int inputIndex, byte[] connectedScript, long value, byte sigHashType) {
        ArrayList<byte[]> byteArray = new ArrayList<>();
        byte sigHashMask = 0x1f;
        int totalLen = 0;

        if (inputIndex > getInputs().size() - 1) {
            throw new RuntimeException("inputIndex out of range");
        }

        totalLen += appendBytes(byteArray, Util.intToBytesLittle((int) getVersion()));

        if ((sigHashType & SigHash.ANYONECANPAY.value) == 0) {
            ArrayList<byte[]> byteArrayTmp = new ArrayList<byte[]>();
            int tmplen = 0;
            for (TransactionInput tmpInput : getInputs()) {
                tmplen += appendBytes(byteArrayTmp, tmpInput.getOutpoint().bitcoinSerialize());
            }
            byte[] tmpBufferBytes = getFinalBytes(byteArrayTmp, tmplen);
            Sha256Hash tmpHash = Sha256Hash.twiceOf(tmpBufferBytes);
            totalLen += appendBytes(byteArray, tmpHash.getBytes());
        } else {
            totalLen += appendBytes(byteArray, Sha256Hash.ZERO_HASH.getBytes());
        }

        if ((sigHashType & SigHash.ANYONECANPAY.value) == 0 &&
                (sigHashType & sigHashMask) != SigHash.SINGLE.value &&
                (sigHashType & sigHashMask) != SigHash.NONE.value) {
            ArrayList<byte[]> byteArrayTmp = new ArrayList<>();
            int tmplen = 0;
            for (TransactionInput tmpInput : getInputs()) {
                tmplen += appendBytes(byteArrayTmp, Util.intToBytesLittle((int) tmpInput.getSequenceNumber()));
            }

            byte[] tmpBufferBytes = getFinalBytes(byteArrayTmp, tmplen);
            Sha256Hash tmpHash = Sha256Hash.twiceOf(tmpBufferBytes);
            totalLen += appendBytes(byteArray, tmpHash.getBytes());
        } else {
            totalLen += appendBytes(byteArray, Sha256Hash.ZERO_HASH.getBytes());
        }

        totalLen += appendBytes(byteArray, getInput(inputIndex).getOutpoint().bitcoinSerialize());
        totalLen += appendVarInt(byteArray, connectedScript);
        totalLen += appendBytes(byteArray, connectedScript);
        totalLen += appendBytes(byteArray, Util.longToBytesLittle(value));
        totalLen += appendBytes(byteArray, Util.intToBytesLittle((int) getInput(inputIndex).getSequenceNumber()));

        if ((sigHashType & SigHash.SINGLE.value) != SigHash.SINGLE.value &&
                (sigHashType & SigHash.NONE.value) != SigHash.NONE.value) {
            ArrayList<byte[]> byteArrayTmp = new ArrayList<byte[]>();
            int tmplen = 0;
            for (TransactionOutput tmpOutput : getOutputs()) {
                tmplen += appendBytes(byteArrayTmp, tmpOutput.bitcoinSerialize());
            }
            byte[] tmpBufferBytes = getFinalBytes(byteArrayTmp, tmplen);
            Sha256Hash tmpHash = Sha256Hash.twiceOf(tmpBufferBytes);
            totalLen += appendBytes(byteArray, tmpHash.getBytes());
        } else if ((sigHashType & sigHashMask) == SigHash.SINGLE.value &&
                inputIndex < getOutputs().size()) {
            byte[] tmpBytes = getOutput(inputIndex).bitcoinSerialize();
            Sha256Hash tmpHash = Sha256Hash.twiceOf(tmpBytes);
            totalLen += appendBytes(byteArray, tmpHash.getBytes());
        } else {
            totalLen += appendBytes(byteArray, Sha256Hash.ZERO_HASH.getBytes());
        }

        totalLen += appendBytes(byteArray, Util.intToBytesLittle((int) getLockTime()));
        totalLen += appendBytes(byteArray, Util.intToBytesLittle(sigHashType | ForkID));
        byte[] finalBytes = getFinalBytes(byteArray, totalLen);
        return Sha256Hash.twiceOf(finalBytes);
    }

    private Script createVinScript(int inputIndex, ECKey key, byte[] connectedScript, long value, SigHash hashType) {
        byte trueSigHashByte = (byte) (hashType.byteValue() | ForkID);
        Sha256Hash hash = this.hashForSignature(inputIndex, connectedScript, value, trueSigHashByte);
        ECKey.ECDSASignature ecdsaSignature = key.sign(hash);
        byte[] ecdsaSignatureByte = ecdsaSignature.encodeToDER();
        byte[] trueSignResult = new byte[ecdsaSignatureByte.length + 1];
        System.arraycopy(ecdsaSignatureByte, 0, trueSignResult, 0, ecdsaSignatureByte.length);
        trueSignResult[trueSignResult.length - 1] = trueSigHashByte;
        return new ScriptBuilder().data(trueSignResult).data(key.getPubKey()).build();
    }

    public String signBSVTransaction(DeterministicKey key, List<SignParam> signParams) throws Exception {
        List<TransactionInput> inputs = this.getInputs();
        if (inputs.size() != signParams.size()) {
            throw new Exception(String.format("transactionInputs size %d , but signParams size %d", inputs.size(), signParams.size()));
        }
        for (int i = 0; i < signParams.size(); i++) {
            int changeStep = 0;
            if (signParams.get(i).isChange()) {
                changeStep = 1;
            }
            DeterministicKey step4 = HDKeyDerivation.deriveChildKey(key, new ChildNumber(changeStep, false));
            DeterministicKey privateKey = HDKeyDerivation.deriveChildKey(step4, new ChildNumber(signParams.get(i).getAddressIndex(), false));
            byte[] preScript = Hex.decode(signParams.get(i).getScriptHex().getBytes());
            Script inputScript = this.createVinScript(i, privateKey, preScript, signParams.get(i).getAmount(), SigHash.ALL);
            this.getInput(i).setScriptSig(inputScript);
        }
        return new String(Hex.encode(this.bitcoinSerialize()));
    }
}
