package cn.sliyun.dotwallet;

import cn.sliyun.dotwallet.entity.DotBaseResponse;
import cn.sliyun.dotwallet.entity.PKIRequest;
import cn.sliyun.dotwallet.entity.Vout;
import cn.sliyun.dotwallet.entity.response.BaseResponse;
import cn.sliyun.dotwallet.entity.response.BroadcastTransactionResponse;
import cn.sliyun.dotwallet.entity.response.BuildTransactionResponse;
import cn.sliyun.dotwallet.entity.response.GetMerkleProofResponse;
import cn.sliyun.dotwallet.util.HttpUtil;
import cn.sliyun.http.OkHttpUtils;
import com.alibaba.fastjson.JSON;
import okhttp3.MediaType;
import okhttp3.Response;
import org.bitcoinj.core.ECKey;
import org.bitcoinj.core.Sha256Hash;
import org.bitcoinj.crypto.ChildNumber;
import org.bitcoinj.crypto.DeterministicKey;
import org.bitcoinj.crypto.HDKeyDerivation;
import org.bitcoinj.crypto.MnemonicCode;
import org.bitcoinj.params.AbstractBitcoinNetParams;
import org.bitcoinj.params.MainNetParams;
import org.spongycastle.util.encoders.Hex;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


public class DotWalletClient {
    private static final String DotWalletHost = "https://api.ddpurse.com";
    public static final AbstractBitcoinNetParams BSVNetParams = MainNetParams.get();
    private DeterministicKey masterKey;

    /**
     * 打点HD钱包使用路径为：m/44'/0'/0'
     *
     * @param masterKey main HDKey
     */
    private DotWalletClient(DeterministicKey masterKey) {
        //path:m/44'
        DeterministicKey step1 = HDKeyDerivation.deriveChildKey(masterKey, new ChildNumber(44, true));
        //path:m/44'/0'
        DeterministicKey step2 = HDKeyDerivation.deriveChildKey(step1, new ChildNumber(0, true));
        //path:m/44'/0'/0'
        DeterministicKey step3 = HDKeyDerivation.deriveChildKey(step2, new ChildNumber(0, true));
        this.masterKey = step3;
    }

    /**
     * @param mnemonicWords 助记词
     * @return
     * @throws Exception
     */
    public static DotWalletClient newInstance(List<String> mnemonicWords) throws Exception {
        if (mnemonicWords == null || mnemonicWords.size() == 0) {
            throw new Exception("mnemonic words empty");
        }
        byte[] seed = MnemonicCode.toSeed(mnemonicWords, "");
        return new DotWalletClient(HDKeyDerivation.createMasterPrivateKey(seed));
    }

    /**
     * 注册xPub
     *
     * @param userIndex 该xpub与用户的关联，不可重复，不可为0,需要记住该数字
     * @param xPub      公钥
     * @return
     * @throws Exception
     */
    public BaseResponse registerXPub(Integer userIndex, String xPub) throws Exception {
        Map<String, Object> params = new HashMap<>();
        params.put("appid", "test_bsv_coin_regular");
        params.put("user_index", userIndex);
        params.put("xpub", xPub);
        params.put("seed", xPub);
        return doPKIRequest("/v1/bsv/register_xpub", params, BaseResponse.class);
    }

    /**
     * 构建未签名交易
     *
     * @param userIndex  注册公钥时填写的userIndex
     * @param vouts      交易输出
     * @param feeByte    交易手续费费率，最终以feesatoshi/feebyte计算，表示每字节多少聪，建议值`2`
     * @param feeSatoshi 交易手续费费率，最终以feesatoshi/feebyte计算，表示每字节多少聪，建议值`2`
     * @return
     * @throws Exception
     */
    public BuildTransactionResponse buildTransaction(Integer userIndex, List<Vout> vouts, Integer feeByte, Integer feeSatoshi) throws Exception {
        Map<String, Object> sender = new HashMap<>();
        sender.put("appid", "test_bsv_coin_regular");
        sender.put("user_index", userIndex);
        Map<String, Object> params = new HashMap<>();
        params.put("sender", sender);
        params.put("vouts", vouts);
        params.put("feebyte", feeByte);
        params.put("feesatoshi", feeSatoshi);
        params.put("sign", false);
        params.put("send", false);
        return doPKIRequest("/v1/bsv/send_to_script", params, BuildTransactionResponse.class);
    }

    /**
     * 广播交易
     *
     * @param rawtx     十六进制已签名的原始交易
     * @param unlimited 是否突破链式交易上线
     * @return
     * @throws Exception
     */
    public BroadcastTransactionResponse broadcastTransaction(String rawtx, Boolean unlimited) throws Exception {
        Map<String, Object> params = new HashMap<>();
        params.put("rawtx", rawtx);
        params.put("unlimit", unlimited);
        return doPKIRequest("/v1/bsv/broadcast_transaction", params, BroadcastTransactionResponse.class);
    }

    /**
     * 获取默克尔证明
     *
     * @param txid 交易txid
     * @return
     * @throws Exception
     */
    public GetMerkleProofResponse getMerkleProof(String txid) throws Exception {
        Map<String, Object> params = new HashMap<>();
        params.put("txid", txid);
        return doPKIRequest("/v1/bsvchain/get_merkleproof", params, GetMerkleProofResponse.class);
    }

    public GetMerkleProofResponse getSubUser() throws Exception {
        Map<String, Object> params = new HashMap<>();
        return doPKIRequest(DotApiConstants.getSubUserList, params, GetMerkleProofResponse.class);
    }


    public DotBaseResponse getSubUserList()  {
        Map<String, Object> params = new HashMap<>();
        return doRequest(DotApiConstants.getSubUserList,params);
    }


    public DotBaseResponse startSolo(int puid, String address) {
        Map<String, Object> params = new HashMap<>();
        params.put("puid", puid);
        params.put("is_solo", 1);
        params.put("address", address);
        return doRequest(DotApiConstants.setMiningType,params);
    }

    public DotBaseResponse stopSolo(int puid, String address) {
        Map<String, Object> params = new HashMap<>();
        params.put("puid", puid);
        params.put("is_solo", 0);
        params.put("address", address);
        return doRequest(DotApiConstants.setMiningType,params);
    }


    public DotBaseResponse getBlockList(long startTime, long endTime)  {
        Map<String, Object> params = new HashMap<>();
        params.put("start_time", startTime);
        params.put("end_time", endTime);
        return doRequest(DotApiConstants.getBlockList,params);
    }


    public DotBaseResponse getShareAccept(long startTime, long endTime)  {
        Map<String, Object> params = new HashMap<>();
        params.put("start_time", startTime);
        params.put("end_time", endTime);
        return doRequest(DotApiConstants.getShareAccept,params);
    }


    private DotBaseResponse doRequest(String requestURL, Map<String, Object> params) {
        String payload = JSON.toJSONString(params);
        //path:m/44'/0'/0'/0
        DeterministicKey step4 = HDKeyDerivation.deriveChildKey(this.masterKey, new ChildNumber(0, false));
        //path:m/44'/0'/0'/0/0
        DeterministicKey PKIKey = HDKeyDerivation.deriveChildKey(step4, new ChildNumber(0, false));
        Sha256Hash hashToSign = Sha256Hash.of(payload.getBytes());
        ECKey.ECDSASignature ecdsaSignature = PKIKey.sign(hashToSign);
        PKIRequest request = new PKIRequest();
        request.setPayload(payload);
        request.setSignature(Hex.toHexString(ecdsaSignature.encodeToDER()));
        request.setPubkey(Hex.toHexString(PKIKey.getPubKey()));
        String json =  JSON.toJSONString(request);
        System.out.println("doRequest json :"+json);
        Response response = null;
        try {
            response =
                    OkHttpUtils
                            .postString()
                            .url(DotWalletHost+requestURL)
                            .content(json)
                            .mediaType(MediaType.parse("application/json; charset=utf-8"))
                            .build()
                            .execute();
            String re = response.body().string();
            System.out.println("response:" + re);
            DotBaseResponse baseResponse = JSON.parseObject(re, DotBaseResponse.class);
            return baseResponse;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    private <T extends BaseResponse> T
    doPKIRequest(String requestURL, Map<String, Object> params, Class<T> tClass) throws Exception {
        String payload = JSON.toJSONString(params);
        //path:m/44'/0'/0'/0
        DeterministicKey step4 = HDKeyDerivation.deriveChildKey(this.masterKey, new ChildNumber(0, false));
        //path:m/44'/0'/0'/0/0
        DeterministicKey PKIKey = HDKeyDerivation.deriveChildKey(step4, new ChildNumber(0, false));
        Sha256Hash hashToSign = Sha256Hash.of(payload.getBytes());
        ECKey.ECDSASignature ecdsaSignature = PKIKey.sign(hashToSign);
        PKIRequest request = new PKIRequest();
        request.setPayload(payload);
        request.setSignature(Hex.toHexString(ecdsaSignature.encodeToDER()));
        request.setPubkey(Hex.toHexString(PKIKey.getPubKey()));
        String json =  JSON.toJSONString(request);
        System.out.println("doPKIRequest json :"+json);

        String result = HttpUtil.httpPostWithJson(DotWalletHost+requestURL, json);
        T response;
        try {
            if (result.equals("")) {
                throw new RuntimeException("empty http result");
            }
            response = JSON.parseObject(result, tClass);
            return response;
        } catch (Exception e) {
            try {
                response = tClass.newInstance();
                Field code = tClass.getSuperclass().getDeclaredField("code");
                code.setAccessible(true);
                code.set(response, -1);
                Field msg = tClass.getSuperclass().getDeclaredField("msg");
                msg.setAccessible(true);
                msg.set(response, result);
                return response;
            } catch (Exception e1) {
                throw new RuntimeException(e1);
            }
        }
    }
}
