package cn.sliyun.framework.web.exception;

import cn.sliyun.common.core.Result;
import cn.sliyun.common.core.ResultGenerator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.authentication.AccountExpiredException;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.validation.BindException;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.validation.ObjectError;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.servlet.NoHandlerFoundException;
import cn.sliyun.common.exception.BaseException;
import cn.sliyun.common.exception.AppException;
import cn.sliyun.common.exception.DemoModeException;
import cn.sliyun.common.utils.StringUtils;
import cn.sliyun.framework.web.domain.AjaxResult;

import java.util.List;
import java.util.concurrent.atomic.AtomicReference;

/**
 * 全局异常处理器
 *
 * @author ruoyi
 */
@RestControllerAdvice
public class GlobalExceptionHandler {
    private static final Logger log = LoggerFactory.getLogger(GlobalExceptionHandler.class);

    /**
     * 基础异常
     */
    @ExceptionHandler(BaseException.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result baseException(BaseException e) {
        return ResultGenerator.genFailResult(e.getMessage());
    }

    /**
     * 业务异常
     */

    @ExceptionHandler(NoHandlerFoundException.class)
    @ResponseStatus(code = HttpStatus.NOT_FOUND)
    public Result handlerNoFoundException(Exception e) {
        log.error(e.getMessage(), e);
        return ResultGenerator.genFailResult("路径不存在，请检查路径是否正确");
    }

    @ExceptionHandler(AccessDeniedException.class)
    @ResponseStatus(code = HttpStatus.FORBIDDEN)
    public Result handleAuthorizationException(AccessDeniedException e) {
        log.error(e.getMessage());
        return ResultGenerator.genFailResult("没有权限，请联系管理员授权");
    }

    @ExceptionHandler(AccountExpiredException.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result handleAccountExpiredException(AccountExpiredException e) {
        log.error(e.getMessage(), e);
        return ResultGenerator.genFailResult(e.getMessage());
    }

    @ExceptionHandler(UsernameNotFoundException.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result handleUsernameNotFoundException(UsernameNotFoundException e) {
        log.error(e.getMessage(), e);
        return ResultGenerator.genFailResult(e.getMessage());
    }

    @ExceptionHandler(Exception.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result handleException(Exception e) {
        log.error(e.getMessage(), e);
        return ResultGenerator.genFailResult(e.getMessage());
    }

    /**
     * 自定义验证异常
     */
    @ExceptionHandler(BindException.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result validatedBindException(BindException e) {
        for (int i = 0; i < e.getFieldErrors().size(); i++) {
            return ResultGenerator.genFailResult(e.getFieldErrors().get(i).getDefaultMessage());
        }
        return ResultGenerator.genFailResult(e.getMessage());
    }

    /**
     * 自定义验证异常
     */
    @ExceptionHandler(MethodArgumentNotValidException.class)
    @ResponseBody
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result exceptionHandler(MethodArgumentNotValidException e) {
        BindingResult bindingResult = e.getBindingResult();
        List<ObjectError> allErrors = bindingResult.getAllErrors();

        AtomicReference<String> error = new AtomicReference<>("");
        allErrors.forEach(objectError -> {
            FieldError fieldError = (FieldError) objectError;
            error.set(fieldError.getField() + "|" + fieldError.getDefaultMessage());
        });
        return ResultGenerator.genFailResult(400, error.get());
    }

    /**
     * 演示模式异常
     */
    @ExceptionHandler(DemoModeException.class)
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result demoModeException(DemoModeException e) {
        return ResultGenerator.genFailResult(400, e.getMessage());
    }

    /**
     * 业务异常
     */
    @ExceptionHandler
    @ResponseBody
    @ResponseStatus(code = HttpStatus.BAD_REQUEST)
    public Result exceptionHandler(AppException e) {
        return ResultGenerator.genFailResult(e.getCode(), e.getMessage());
    }

}
